﻿#pragma once
#include "span.h"
#include "metadata_allocator.h"
#define PTR_TO_LV0(ptr)	(ptr) >> 25
#define PTR_TO_LV1(ptr)	((ptr) >> 15 & 0x3FF)
#define PTR_TO_LV2(ptr)	((ptr) & 0x7FFF)
namespace bg
{
    namespace detail
    {
        template<size_t size0, size_t size1, size_t size2>
        struct RadixTree
        {

            struct  NodeV2
            {
                Span* lv2[1 << size2]{0};
            };
            struct NodeV1
            {
                NodeV2* lv1[1 << size1]{0};
            };

            NodeV1* lv0[1 << size0]{0};
        };

        class PageHeap
        {
        public:
            PageHeap();

            bg::detail::Span* GetSpanMap(size_t start_page);

            bg::detail::Span* AllocateSpan(size_t page_count);

            void DeallocateSpan(bg::detail::Span* s);

            void RegisterSpan(bg::detail::Span* s);
        private:

            bool SetSpanMap(size_t page, bg::detail::Span* span);

            bool GrowHeap(size_t page_count);

            bg::detail::Span* GetSpanByFreeList(size_t page_count);

            bg::detail::Span* GetSpanByLargeList(size_t page_count);

            bg::detail::Span* GetListSpan(size_t page_count)
            {
                Span* result = nullptr;
                result = GetSpanByFreeList(page_count);
                if(!result)
                {
                    result = GetSpanByLargeList(page_count);
                }
                return result;
            }

            using  RadixTree = RadixTree<10ul, 10ul, 15ul>;
        private:

            static constexpr size_t MAX_FREE_LIST_COUNT = 128;
            static constexpr size_t DEFULT_ALLOC_BYTES = PAGES_TO_BYTES(MAX_FREE_LIST_COUNT);

            RadixTree m_span_map;
            MetadataAllocator<bg::detail::Span> m_span_allocator;
            Span m_free_lists[128]{};
            Span m_large_list{};
        };
    }
}