﻿#pragma once
#include "shm_config.h"
#include "shm_helper.h"

#define BIG_CALIGNMENT 15488                                   // 大字节转化系数 
#define SMALL_CALIGNMENT 8			                           // 小字节转化系数
#define SMALL_BYTES_TO_INDEX(bytes) ((bytes + SMALL_CALIGNMENT - 1) >> 3)        // 小字节获取class index
#define BIG_BYTES_TO_INDEX(bytes) ((bytes + (BIG_CALIGNMENT - 1)) >> 7)         // 大字节获取class index
#define BYTES_COMP_VALUE 0x400                                  // 对比大小判断数
#define CLASS_MAX_COUNT 87									    // 总共类数量    

namespace bg
{
    namespace detail
    {
        class SizeMap
        {
        public:
            /// @brief 初始化CLASS_MAX_BYTES以下内存分页配置
            bool Init();

            size_t GetClassByBytes(size_t bytes)
            {
                if(bytes <= BYTES_COMP_VALUE)
                {
                    return m_index_to_class[SMALL_BYTES_TO_INDEX(bytes)];
                }
                return m_index_to_class[BIG_BYTES_TO_INDEX(bytes)];
            }

            size_t ClassToSize(size_t index)
            {
                SHM_ASSERT_RETURN_VALUE(index < CLASS_MAX_COUNT, 0);
                return m_class_to_size[index];
            }

            size_t ClassToPages(size_t index)
            {
                SHM_ASSERT_RETURN_VALUE(index < CLASS_MAX_COUNT, 0);
                return m_class_to_pages[index];
            }

            static size_t CalcAlignment(size_t bytes);

        private:
            uint8_t m_index_to_class[2176]{0};
            size_t m_class_to_size[CLASS_MAX_COUNT]{0};
            size_t m_class_to_pages[CLASS_MAX_COUNT]{0};
        };


    }
}