
#pragma once
#include <stdint.h>
#include "shm_helper.h"
namespace bg
{
    namespace detail
    {
        struct Span
        {
            Span() = default;
            Span(size_t start_page, size_t page_count) : m_start_page(start_page), m_page_count(page_count) {}

            void RemoveNextSpan()
            {
                auto next_span = m_next;

                m_next = next_span->m_next;
                next_span->m_next->m_prev = next_span->m_prev;

                next_span->m_prev = nullptr;
                next_span->m_next = nullptr;
            }

            void InstertNextSpan(Span* span)
            {
                span->m_prev = m_next->m_prev;
                span->m_next = m_next;
                m_next->m_prev = span;
                m_next = span;
            }

            void RemoveSelfForList()
            {
                m_prev->RemoveNextSpan();
            }

            bool GetUseState() { return m_in_use; }

            void SetUseState(bool state) { m_in_use = state; }

            uint8_t GetSizeClass() { return m_size_class; }

            void SetSizeClass(uint8_t size_class) { m_size_class = size_class; }

            size_t GetUsedCount()
            {
                return m_used_count;
            }

            void IncUsedCount()
            {
                m_used_count++;
            }

            void DecUsedCout()
            {
                SHM_ASSERT_RETURN_VOID(m_used_count);
                m_used_count--;
            }

            void SetUsedCount(size_t used_count)
            {
                SHM_ASSERT_RETURN_VOID(used_count >= 0);
                m_used_count = used_count;
            }

            void SetStartPage(size_t start_page)
            {
                m_start_page = start_page;
            }

            size_t GetStartPage()
            {
                return m_start_page;
            }

            size_t GetLastPage()
            {
                SHM_ASSERT_RETURN_VALUE(m_page_count > 0, m_start_page);
                return m_start_page + m_page_count - 1;
            }

            size_t GetPageCount()
            {
                return m_page_count;
            }

            void SetPageCount(size_t page_count)
            {
                SHM_ASSERT_RETURN_VOID(page_count >= 0);
                m_page_count = page_count;
            }

            Span* GetNextSpan()
            {
                return m_next;
            }

            Span* GetPreSpan()
            {
                return m_next;
            }

            void SetChunkList(void* chunk_list)
            {
                m_chunk_list = chunk_list;
            }
            void* GetChunkList() { return m_chunk_list; }

            void LeftMovePages(size_t page_count)
            {
                m_start_page -= page_count;
                m_page_count += page_count;
            }

            void RightMovePages(size_t page_count)
            {
                m_page_count += page_count;
            }

        private:


            bool m_in_use{};
            uint8_t m_size_class{0xff};
            size_t m_used_count{};
            size_t m_start_page{};
            size_t m_page_count{};

            Span* m_prev{this};
            Span* m_next{this};
            void* m_chunk_list{};

        };
    }
}